@@ -42,16 +42,24 @@ enum q931_mode;
 
 struct pri {
 	int fd;				/* File descriptor for D-Channel */
+	struct pri *subchannel;	/* Sub-channel if appropriate */
+	struct pri *master;		/* Master channel if appropriate */
 	struct pri_sched pri_sched[MAX_SCHED];	/* Scheduled events */
 	int debug;			/* Debug stuff */
 	int state;			/* State of D-channel */
 	int switchtype;		/* Switch type */
+	int nsf;		/* Network-Specific Facility (if any) */
 	int localtype;		/* Local network type (unknown, network, cpe) */
 	int remotetype;		/* Remote network type (unknown, network, cpe) */
+
+	int sapi;
+	int tei;
+	int protodisc;
 	
 	/* Q.921 State */
 	int q921_state;	
 	int window;			/* Max window size */
+	int windowlen;		/* Fullness of window */
 	int v_s;			/* Next N(S) for transmission */
 	int v_a;			/* Last acknowledged frame */
 	int v_r;			/* Next frame expected to be received */
@@ -78,12 +86,40 @@ struct pri {
 	struct q921_frame *txqueue;
 	
 	/* Q.931 calls */
-	q931_call *calls;
+	q931_call **callpool;
+	q931_call *localpool;
 
 	/* do we do overlap dialing */
 	int overlapdial;
+
+#ifdef LIBPRI_COUNTERS
+	/* q921/q931 packet counters */
+	unsigned int q921_txcount;
+	unsigned int q921_rxcount;
+	unsigned int q931_txcount;
+	unsigned int q931_rxcount;
+#endif
 };
 
+struct pri_sr {
+	int transmode;
+	int channel;
+	int exclusive;
+	int nonisdn;
+	char *caller;
+	int callerplan;
+	char *callername;
+	int callerpres;
+	char *called;
+	int calledplan;
+	int userl1;
+	int numcomplete;
+};
+
+/* Internal switch types */
+#define PRI_SWITCH_GR303_EOC_PATH	10
+#define PRI_SWITCH_GR303_TMC_SWITCHING	11
+
 extern int pri_schedule_event(struct pri *pri, int ms, void (*function)(void *data), void *data);
 
 extern pri_event *pri_schedule_run(struct pri *pri);
