@@ -26,7 +26,14 @@
 #define _PRI_Q921_H
 
 #include <sys/types.h>
+#if defined(__linux__)
 #include <endian.h>
+#elif defined(__FreeBSD__)
+#include <sys/endian.h>
+#define __BYTE_ORDER _BYTE_ORDER
+#define __BIG_ENDIAN _BIG_ENDIAN
+#define __LITTLE_ENDIAN _LITTLE_ENDIAN
+#endif
 
 /* Timer values */
 
@@ -43,8 +50,17 @@
 #define Q921_FRAMETYPE_S	0x1
 
 #define Q921_TEI_GROUP				127
+#define Q921_TEI_GR303_EOC_PATH			0
+#define Q921_TEI_GR303_EOC_OPS			4
+#define Q921_TEI_GR303_TMC_SWITCHING		0
+#define Q921_TEI_GR303_TMC_CALLPROC		0
+
+#define Q921_SAPI_CALL_CTRL		0
+#define Q921_SAPI_GR303_EOC		1
+#define Q921_SAPI_GR303_TMC_SWITCHING	1
+#define Q921_SAPI_GR303_TMC_CALLPROC	0
+
 
-#define Q921_SAPI_CALL_CTRL			0
 #define Q921_SAPI_PACKET_MODE		1
 #define Q921_SAPI_X25_LAYER3      	16
 #define Q921_SAPI_LAYER2_MANAGEMENT	63
@@ -134,6 +150,7 @@ typedef union {
 typedef struct q921_frame {
 	struct q921_frame *next;	/* Next in list */
 	int len;					/* Length of header + body */
+	int transmitted;			/* Have we been transmitted */
 	q921_i h;
 } q921_frame;
 
