@@ -41,12 +41,12 @@
 #define RANDOM_DROPS
 */
 
-#define Q921_INIT(hf) do { \
+#define Q921_INIT(pri, hf) do { \
 	memset(&(hf),0,sizeof(hf)); \
-	(hf).h.sapi = 0; \
+	(hf).h.sapi = (pri)->sapi; \
 	(hf).h.ea1 = 0; \
 	(hf).h.ea2 = 1; \
-	(hf).h.tei = 0; \
+	(hf).h.tei = (pri)->tei; \
 } while(0)
 
 static void reschedule_t203(struct pri *pri);
@@ -67,13 +67,17 @@ static void q921_discard_retransmissions(struct pri *pri)
 static int q921_transmit(struct pri *pri, q921_h *h, int len) 
 {
 	int res;
+	if (pri->master)
+		return q921_transmit(pri->master, h, len);
 #ifdef RANDOM_DROPS
    if (!(random() % 3)) {
          pri_message(" === Dropping Packet ===\n");
          return 0;
    }
 #endif   
-      
+#ifdef LIBPRI_COUNTERS
+	pri->q921_txcount++;      
+#endif
 	/* Just send it raw */
 	if (pri->debug & PRI_DEBUG_Q921_DUMP)
 		q921_dump(h, len, pri->debug & PRI_DEBUG_Q921_RAW, 1);
@@ -90,7 +94,7 @@ static int q921_transmit(struct pri *pri, q921_h *h, int len)
 static void q921_send_ua(struct pri *pri, int pfbit)
 {
 	q921_h h;
-	Q921_INIT(h);
+	Q921_INIT(pri, h);
 	h.u.m3 = 3;		/* M3 = 3 */
 	h.u.m2 = 0;		/* M2 = 0 */
 	h.u.p_f = pfbit;	/* Final bit on */
@@ -122,7 +126,7 @@ static void q921_send_sabme(void *vpri, int now)
 	pri->sabme_timer = pri_schedule_event(pri, T_200, q921_send_sabme_now, pri);
 	if (!now)
 		return;
-	Q921_INIT(h);
+	Q921_INIT(pri, h);
 	h.u.m3 = 3;	/* M3 = 3 */
 	h.u.m2 = 3;	/* M2 = 3 */
 	h.u.p_f = 1;	/* Poll bit set */
@@ -168,6 +172,23 @@ static int q921_ack_packet(struct pri *pri, int num)
 			free(f);
 			/* Reset retransmission counter if we actually acked something */
 			pri->retrans = 0;
+			/* Decrement window size */
+			pri->windowlen--;
+			/* Search for something to send */
+			f = pri->txqueue;
+			while(f) {
+				if (!f->transmitted) {
+					/* Send it now... */
+					if (pri->debug & PRI_DEBUG_Q921_STATE)
+						pri_message("-- Finally transmitting %d, since window opened up\n", f->h.n_s);
+					f->transmitted++;
+					pri->windowlen++;
+					f->h.n_r = pri->v_r;
+					q921_transmit(pri, (q921_h *)(&f->h), f->len);
+					break;
+				}
+				f = f->next;
+			}
 			return 1;
 		}
 		prev = f;
@@ -200,7 +221,7 @@ static pri_event *q921_ack_rx(struct pri *pri, int ack)
 	for (x=pri->v_a; (x != pri->v_s) && (x != ack); Q921_INC(x));
 	if (x != ack) {
 		/* ACK was outside of our window --- ignore */
-		pri_error("ACK received outside of window, restarting\n");
+		pri_error("ACK received for '%d' outside of window of '%d' to '%d', restarting\n", ack, pri->v_a, pri->v_s);
 		ev = q921_dchannel_down(pri);
 		q921_start(pri, 1);
 		pri->schedev = 1;
@@ -242,7 +263,7 @@ static pri_event *q921_ack_rx(struct pri *pri, int ack)
 static void q921_reject(struct pri *pri, int pf)
 {
 	q921_h h;
-	Q921_INIT(h);
+	Q921_INIT(pri, h);
 	h.s.x0 = 0;	/* Always 0 */
 	h.s.ss = 2;	/* Reject */
 	h.s.ft = 1;	/* Frametype (01) */
@@ -267,7 +288,7 @@ static void q921_reject(struct pri *pri, int pf)
 
 static void q921_rr(struct pri *pri, int pbit, int cmd) {
 	q921_h h;
-	Q921_INIT(h);
+	Q921_INIT(pri, h);
 	h.s.x0 = 0;	/* Always 0 */
 	h.s.ss = 0; /* Receive Ready */
 	h.s.ft = 1;	/* Frametype (01) */
@@ -317,8 +338,11 @@ static void t200_expire(void *vpri)
             pri_message("-- Retransmitting %d bytes\n", pri->txqueue->len);
 		if (pri->busy) 
 			q921_rr(pri, 1, 0);
-		else
+		else {
+			if (!pri->txqueue->transmitted) 
+				pri_error("!! Not good - head of queue has not been transmitted yet\n");
 			q921_transmit(pri, (q921_h *)&pri->txqueue->h, pri->txqueue->len);
+		}
          if (pri->debug & PRI_DEBUG_Q921_STATE) 
                pri_message("-- Rescheduling retransmission (%d)\n", pri->retrans);
          pri->t200_timer = pri_schedule_event(pri, T_200, t200_expire, pri);
@@ -331,9 +355,26 @@ static void t200_expire(void *vpri)
          q921_start(pri, 1);
          pri->schedev = 1;
       }
+	} else if (pri->solicitfbit) {
+         if (pri->debug & PRI_DEBUG_Q921_STATE)
+            pri_message("-- Retrying poll with f-bit\n");
+		pri->retrans++;
+		if (pri->retrans < N_200) {
+			pri->solicitfbit = 1;
+			q921_rr(pri, 1, 1);
+			pri->t200_timer = pri_schedule_event(pri, T_200, t200_expire, pri);
+		} else {
+			if (pri->debug & PRI_DEBUG_Q921_STATE) 
+				pri_message("-- Timeout occured, restarting PRI\n");
+			pri->q921_state = Q921_LINK_CONNECTION_RELEASED;
+			pri->t200_timer = 0;
+			q921_dchannel_down(pri);
+			q921_start(pri, 1);
+			pri->schedev = 1;
+		}
 	} else {
 		pri_error("T200 counter expired, nothing to send...\n");
-   	pri->t200_timer = 0;
+	   	pri->t200_timer = 0;
 	}
 }
 
@@ -342,9 +383,9 @@ int q921_transmit_iframe(struct pri *pri, void *buf, int len, int cr)
 	q921_frame *f, *prev=NULL;
 	for (f=pri->txqueue; f; f = f->next) prev = f;
 	f = malloc(sizeof(q921_frame) + len + 2);
-	memset(f,0,sizeof(q921_frame) + len + 2);
 	if (f) {
-		Q921_INIT(f->h);
+		memset(f,0,sizeof(q921_frame) + len + 2);
+		Q921_INIT(pri, f->h);
 		switch(pri->localtype) {
 		case PRI_NETWORK:
 			if (cr)
@@ -360,6 +401,7 @@ int q921_transmit_iframe(struct pri *pri, void *buf, int len, int cr)
 		break;
 		}
 		f->next = NULL;
+		f->transmitted = 0;
 		f->len = len + 4;
 		memcpy(f->h.data, buf, len);
 		f->h.n_s = pri->v_s;
@@ -372,9 +414,18 @@ int q921_transmit_iframe(struct pri *pri, void *buf, int len, int cr)
 			prev->next = f;
 		else
 			pri->txqueue = f;
-		/* Immediately transmit unless we're in a recovery state */
+		/* Immediately transmit unless we're in a recovery state, or the window
+		   size is too big */
 		if (!pri->retrans && !pri->busy) {
-			q921_transmit(pri, (q921_h *)(&f->h), f->len);
+			if (pri->windowlen < pri->window) {
+				pri->windowlen++;
+				q921_transmit(pri, (q921_h *)(&f->h), f->len);
+				f->transmitted++;
+			} else {
+				if (pri->debug & PRI_DEBUG_Q921_STATE)
+					pri_message("Delaying transmission of %d, window is %d/%d long\n", 
+						f->h.n_s, pri->windowlen, pri->window);
+			}
 		}
 		if (pri->t203_timer) {
 			if (pri->debug & PRI_DEBUG_Q921_STATE)
@@ -405,12 +456,13 @@ static void t203_expire(void *vpri)
 			pri_message("T203 counter expired, sending RR and scheduling T203 again\n");
 		/* Solicit an F-bit in the other's RR */
 		pri->solicitfbit = 1;
+		pri->retrans = 0;
 		q921_rr(pri, 1, 1);
-		/* Restart ourselves */
-		pri->t203_timer = pri_schedule_event(pri, T_203, t203_expire, pri);
+		/* Start timer T200 to resend our RR if we don't get it */
+		pri->t203_timer = pri_schedule_event(pri, T_200, t200_expire, pri);
 	} else {
 		if (pri->debug &  PRI_DEBUG_Q921_STATE)
-			pri_message("T203 counter expired in weird statd %d\n", pri->q921_state);
+			pri_message("T203 counter expired in weird state %d\n", pri->q921_state);
 		pri->t203_timer = 0;
 	}
 }
@@ -463,10 +515,14 @@ void q921_dump(q921_h *h, int len, int showraw, int txrx)
 	
 	direction_tag = txrx ? '>' : '<';
 	if (showraw) {
-		pri_message("\n%c [", direction_tag);
-		for (x=0;x<len;x++) 
-			pri_message("%02x ",h->raw[x]);
-		pri_message("]");
+		char *buf = malloc(len * 3 + 1);
+		int buflen = 0;
+		if (buf) {
+			for (x=0;x<len;x++) 
+				buflen += sprintf(buf + buflen, "%02x ", h->raw[x]);
+			pri_message("\n%c [ %s]\n", direction_tag, buf);
+			free(buf);
+		}
 	}
 
 	switch (h->h.data[0] & Q921_FRAMETYPE_MASK) {
@@ -625,6 +681,7 @@ void q921_reset(struct pri *pri)
 	pri->v_r = 0;
 	pri->v_na = 0;
 	pri->window = 7;
+	pri->windowlen = 0;
 	pri_schedule_del(pri, pri->sabme_timer);
 	pri_schedule_del(pri, pri->t203_timer);
 	pri_schedule_del(pri, pri->t200_timer);
@@ -641,28 +698,11 @@ void q921_reset(struct pri *pri)
 	q921_discard_retransmissions(pri);
 }
 
-static pri_event *__q921_receive(struct pri *pri, q921_h *h, int len)
+static pri_event *__q921_receive_qualified(struct pri *pri, q921_h *h, int len)
 {
 	q921_frame *f;
 	pri_event *ev;
 	int sendnow;
-	/* Discard FCS */
-	len -= 2;
-	
-	if (pri->debug & PRI_DEBUG_Q921_DUMP)
-		q921_dump(h, len, pri->debug & PRI_DEBUG_Q921_RAW, 0);
-
-	/* Check some reject conditions -- Start by rejecting improper ea's */
-	if (h->h.ea1 || !(h->h.ea2))
-		return NULL;
-
-	/* Check for broadcasts - not yet handled */
-	if (h->h.tei == Q921_TEI_GROUP)
-		return NULL;
-	
-	/* Check for SAPIs we don't yet handle */
-	if (h->h.sapi != Q921_SAPI_CALL_CTRL)
-		return NULL;
 
 	switch(h->h.data[0] & Q921_FRAMETYPE_MASK) {
 	case 0:
@@ -729,8 +769,9 @@ static pri_event *__q921_receive(struct pri *pri, q921_h *h, int len)
 		 sendnow = 0;
          /* Resend the proper I-frame */
          for(f=pri->txqueue;f;f=f->next) {
-               if (sendnow || (f->h.n_s == h->s.n_r)) {
-                     /* Matches the request, or follows in our window */
+               if ((sendnow || (f->h.n_s == h->s.n_r)) && f->transmitted) {
+                     /* Matches the request, or follows in our window, and has
+					    already been transmitted. */
 					 sendnow = 1;
 					 pri_error("!! Got reject for frame %d, retransmitting frame %d now, updating n_r!\n", h->s.n_r, f->h.n_s);
 				     f->h.n_r = pri->v_r;
@@ -853,11 +894,46 @@ static pri_event *__q921_receive(struct pri *pri, q921_h *h, int len)
 	return NULL;
 }
 
+static pri_event *__q921_receive(struct pri *pri, q921_h *h, int len)
+{
+	pri_event *ev;
+	/* Discard FCS */
+	len -= 2;
+	
+	if (!pri->master && pri->debug & PRI_DEBUG_Q921_DUMP)
+		q921_dump(h, len, pri->debug & PRI_DEBUG_Q921_RAW, 0);
+
+	/* Check some reject conditions -- Start by rejecting improper ea's */
+	if (h->h.ea1 || !(h->h.ea2))
+		return NULL;
+
+	/* Check for broadcasts - not yet handled */
+	if (h->h.tei == Q921_TEI_GROUP)
+		return NULL;
+
+	/* Check for SAPIs we don't yet handle */
+	if ((h->h.sapi != pri->sapi) || (h->h.tei != pri->tei)) {
+#ifdef PROCESS_SUBCHANNELS
+		/* If it's not us, try any subchannels we have */
+		if (pri->subchannel)
+			return q921_receive(pri->subchannel, h, len + 2);
+		else 
+#endif
+			return NULL;
+
+	}
+	ev = __q921_receive_qualified(pri, h, len);
+	reschedule_t203(pri);
+	return ev;
+}
+
 pri_event *q921_receive(struct pri *pri, q921_h *h, int len)
 {
 	pri_event *e;
 	e = __q921_receive(pri, h, len);
-	reschedule_t203(pri);
+#ifdef LIBPRI_COUNTERS
+	pri->q921_rxcount++;
+#endif
 	return e;
 }
 
